//-------------------------------------------------------------------------------------------------
// TheApplication.cpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#include "pch.h"
#include "TheApplication.h"
#include "TheConsole.h"
#include "TheMessageHandler.h"
#include "TheVideoSystem.h"
#include "TheRenderer.h"
#include "TheNetwork.h"
#include "TheGameInterfaceAbstract.h"
#include "TheSoundSystem.h"
#include "Timer.h"
#include "engineAssert.h"
using namespace std;


//-------------------------------------------------------------------------------------------------
// Static member initialization
//-------------------------------------------------------------------------------------------------
STATIC TheApplication*	TheApplication::_theApplication		= NULL;
STATIC bool				TheApplication::_isConstructed		= false;
STATIC bool				TheApplication::_isInitialized		= false;
STATIC string			TheApplication::_appName( "UNNAMED APPLICATION : Call APPLICATION->SetName() from PreInit() to set name!" );
//STATIC string			TheApplication::_baseDir( "./" );


//-------------------------------------------------------------------------------------------------
// constructor
//-------------------------------------------------------------------------------------------------
TheApplication::TheApplication( const HINSTANCE& hInstance, const LPSTR& lpCmdLine )
:
_cmdLine( lpCmdLine ),
_hInstance( hInstance ),
_isExiting( false ),
_isShutdownRequested( false ),
_frameNumber( 0 ),
_secondsSinceLastFrame( 0.1f ),
_timeFrameBegan( Timer::GetSysTimeSeconds() ),
_fpsAsymptotic( 60.0f ),
_hasFocus( true )
{
	/// Make sure the singleton object doesn't get constructed twice
	VALIDATE( !_isConstructed );

	/// Set construction complete
	_isConstructed = true;
}


//-------------------------------------------------------------------------------------------------
// destructor
//-------------------------------------------------------------------------------------------------
TheApplication::~TheApplication()
{
	VALIDATE( !_isInitialized && _isConstructed );

	/// Destroy subsystems
	TheNetwork::DestroySingleton();
	TheRenderer::DestroySingleton();
//	TheSoundSystem::DestroySingleton();
	TheVideoSystem::DestroySingleton();
	TheMessageHandler::DestroySingleton();
	TheInputSystem::DestroySingleton();
	TheConsole::DestroySingleton();

	/// Set un-constructed
	_isConstructed = false;
}


//-------------------------------------------------------------------------------------------------
// CreateSingleton
//-------------------------------------------------------------------------------------------------
STATIC TheApplication* TheApplication::CreateSingleton( const HINSTANCE& hInstance, const LPSTR& lpCmdLine )
{
	VALIDATE( !_isConstructed );
	VALIDATE( !_isInitialized );
	VALIDATE( !_theApplication );

	/// Create the singleton and set the static pointer
	_theApplication = new TheApplication( hInstance, lpCmdLine );

	return _theApplication;
}


//-------------------------------------------------------------------------------------------------
// Singleton
//-------------------------------------------------------------------------------------------------
STATIC TheApplication* TheApplication::Singleton()
{
	ASSERT( _isConstructed );
	ASSERT( _theApplication );

	return _theApplication;
}


//-------------------------------------------------------------------------------------------------
// DestroySingleton
//-------------------------------------------------------------------------------------------------
STATIC void TheApplication::DestroySingleton()
{
	VALIDATE_MSG0( _isConstructed, "DestroySingleton() : singleton not constructed!" );
	VALIDATE_MSG0( !_isInitialized, "DestroySingleton() : Must call RequestShutdown() first!" );
	VALIDATE_MSG0( _theApplication, "DestroySingleton() : singleton is NULL!" );

	/// Destroy the singleton and null the static pointer
	delete _theApplication;
	_theApplication = NULL;
}


//-------------------------------------------------------------------------------------------------
// Init
//-------------------------------------------------------------------------------------------------
void TheApplication::Init()
{
//	CONSOLE->Printf( "::: Initialization : Application (starting)\n" );

	/// Make sure the singleton object doesn't get initialized twice
	VALIDATE( _isConstructed );
	VALIDATE( !_isInitialized );
	VALIDATE( _theApplication );

	/// Make sure an instance of the game already exists
	VALIDATE_MSG0( TheGameInterfaceAbstract::IsGamePresent(),
		"ENGINE USAGE ERROR:\n"
		"  You must create a static instance of a class derived from TheGameInterfaceAbstract.\n"
		"  This derived class - typically named TheGameInterface - serves as the main interface"
		"    by which the engine can call into game code.\n"
		"  See TheGameInterfaceAbstract.h and engine documentation for more details.\n" );

	/// Process command line parameters
	_ProcessCommandLine();

	/// Instantiate subsystems
	TheConsole*			theConsole			= TheConsole::CreateSingleton();
	TheInputSystem*		theInputSystem		= TheInputSystem::CreateSingleton();
	TheMessageHandler*	theMessageHandler	= TheMessageHandler::CreateSingleton();
	TheVideoSystem*		theVideoSystem		= TheVideoSystem::CreateSingleton();
	TheSoundSystem*		theSoundSystem		= TheSoundSystem::CreateSingleton();
	TheRenderer*		theRenderer			= TheRenderer::CreateSingleton();
	TheNetwork*			theNetwork			= TheNetwork::CreateSingleton();

	/// Set application initialization complete
	_isInitialized = true;

	/// Inform the game that subsystem initialization is about to begin
	GAMEINTERFACE->PreInit();

	/// Initialize subsystems
	theConsole->Init();
	theInputSystem->Init();
	theMessageHandler->Init();
	theVideoSystem->Init();
	theSoundSystem->Init();
	theRenderer->Init();
	theNetwork->Init();

	/// Inform the game that subsystem initialization has finished
	GAMEINTERFACE->PostInit();

	CONSOLE->Printf( "::: Initialization : Application (finished)\n" );
}


//-------------------------------------------------------------------------------------------------
// _Shutdown
//-------------------------------------------------------------------------------------------------
void TheApplication::_Shutdown()
{
	CONSOLE->Printf( "::: _Shutdown : Application (starting)\n" );

	ASSERT( this );
	ASSERT( _isConstructed );

	/// Enter "shutting down" mode
	_isExiting = true;

	/// Shut down the game
	GAMEINTERFACE->Shutdown();

	/// Shut down subsystems
	NETWORK->Shutdown();
	RENDERER->Shutdown();
	SOUND->Shutdown();
	VIDEO->Shutdown();
	MESSAGE->Shutdown();
	INPUT->Shutdown();
	CONSOLE->Shutdown();

	/// De-initialize the system
	_isInitialized = false;
//	CONSOLE->Printf( "::: _Shutdown : Application (finished)\n" );
}


//-------------------------------------------------------------------------------------------------
// CheckForShutdownRequest
//-------------------------------------------------------------------------------------------------
void TheApplication::CheckForShutdownRequest()
{
	if( _isShutdownRequested )
	{
		_Shutdown();
	}
}


//-------------------------------------------------------------------------------------------------
// RequestShutdown
//-------------------------------------------------------------------------------------------------
void TheApplication::RequestShutdown()
{
	_isShutdownRequested = true;
}


//-------------------------------------------------------------------------------------------------
// OnQuitMessage
//-------------------------------------------------------------------------------------------------
void TheApplication::OnQuitMessage()
{
	ASSERT( this );
	ASSERT( _isConstructed );
	ASSERT( _isInitialized );

	RequestShutdown();
}


//-------------------------------------------------------------------------------------------------
// RunFrame
//-------------------------------------------------------------------------------------------------
void TheApplication::RunFrame()
{
	ASSERT( this );
	ASSERT( _isConstructed );
	ASSERT( _isInitialized );

	/// Don't do any work if a shutdown is in progress
	if( IsExiting() )
		return;

	/// Update the frame counter, etc.
	_UpdateFrameInfo();

	/// Update the sound engine for this frame
	SOUND->Update();

	/// Update the input for this frame
	INPUT->Update();

	/// Tell the game to run (if it exists)
	GAMEINTERFACE->RunFrame();

	/// Tell the renderer to draw everything
	RENDERER->RenderScene();

	/// Swap front / back drawing buffers
	VIDEO->SwapBuffers();

	/// Reset the screen
	RENDERER->ClearScreen( Rgba::BLACK );
	RENDERER->ResetView();

	/// Check if there have been any shutdown requests
	CheckForShutdownRequest();
}


//-------------------------------------------------------------------------------------------------
// _ProcessCommandLine
//-------------------------------------------------------------------------------------------------
void TheApplication::_ProcessCommandLine()
{
	/// Allow the game first pass at the command line; let it change the string if it wants
	GAMEINTERFACE->ProcessCommandLine( _cmdLine );
}



//-------------------------------------------------------------------------------------------------
// _UpdateFrameInfo
//-------------------------------------------------------------------------------------------------
void TheApplication::_UpdateFrameInfo()
{
	/// Advance the frame counter
	_frameNumber ++;

	/// Calculate elapsed time for the previous frame
	double timeNow = Timer::GetSysTimeSeconds();
	_secondsSinceLastFrame = timeNow - _timeFrameBegan;
	_timeFrameBegan = timeNow;

	/// Clamp the time to a reasonable, nonzero range
	_secondsSinceLastFrame = ClampDouble( _secondsSinceLastFrame, 0.001, 1.0 );

	/// Calculate instantaneous framerate
	float fpsInstantaneous = 1.0f / (float) _secondsSinceLastFrame;
	_fpsAsymptotic *= 0.90f;
	_fpsAsymptotic += (0.10f * fpsInstantaneous);
}


//-------------------------------------------------------------------------------------------------
// OnActivate
//-------------------------------------------------------------------------------------------------
void TheApplication::OnAppGainFocus()
{
	_hasFocus = true;

	/// Tell the input system to override mouse cursor hide + lock
	INPUT->OnAppGainFocus();
}


//-------------------------------------------------------------------------------------------------
// OnDeactivate
//-------------------------------------------------------------------------------------------------
void TheApplication::OnAppLoseFocus()
{
	_hasFocus = false;

	/// Tell the input system to remove mouse cursor hide/lock override
	INPUT->OnAppLoseFocus();
}


