//-------------------------------------------------------------------------------------------------
// TheNetwork.cpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#include "pch.h"
#include "TheNetwork.h"
#include "TheConsole.h"
#include "engineAssert.h"
using namespace std;


//-------------------------------------------------------------------------------------------------
// Static member initialization
//-------------------------------------------------------------------------------------------------
STATIC TheNetwork*			TheNetwork::_TheNetwork = NULL;
STATIC bool					TheNetwork::_isConstructed = false;
STATIC bool					TheNetwork::_isInitialized = false;
STATIC CRITICAL_SECTION		TheNetwork::_criticalSection_connectionList;
STATIC unsigned int			TheNetwork::_nextConnectionID = 1;



//-------------------------------------------------------------------------------------------------
// constructor
//-------------------------------------------------------------------------------------------------
TheNetwork::TheNetwork()
:
_isNetworkServiceStarted( false ),
_portNumber( DEFAULT_PORT_NUMBER ),
_maxNumberOfConnections( DEFAULT_MAX_CONNECTIONS ),
_localHostName( "(unknown)" ),
_localHostInfo( NULL ),
_listenSocket( NULL )
{
	/// Make sure the singleton object doesn't get constructed twice
	VALIDATE( !_isConstructed );

	/// Set construction complete
	_isConstructed = true;
}


//-------------------------------------------------------------------------------------------------
// destructor
//-------------------------------------------------------------------------------------------------
TheNetwork::~TheNetwork()
{
	VALIDATE( !_isInitialized && _isConstructed );

	/// Set un-constructed
	_isConstructed = false;
}


//-------------------------------------------------------------------------------------------------
// CreateSingleton
//-------------------------------------------------------------------------------------------------
STATIC TheNetwork* TheNetwork::CreateSingleton()
{
	VALIDATE( !_isConstructed );
	VALIDATE( !_isInitialized );
	VALIDATE( !_TheNetwork );

	/// Create the singleton and set the static pointer
	_TheNetwork = new TheNetwork;

	return _TheNetwork;
}


//-------------------------------------------------------------------------------------------------
// Singleton
//-------------------------------------------------------------------------------------------------
STATIC TheNetwork* TheNetwork::Singleton()
{
	ASSERT( _isConstructed );
	ASSERT( _TheNetwork );

	/// Return NULL if the system is not active or uninitialized
//	if( !_isInitialized )
//	{
//		return NULL;
//	}

	return _TheNetwork;
}


//-------------------------------------------------------------------------------------------------
// DestroySingleton
//-------------------------------------------------------------------------------------------------
STATIC void TheNetwork::DestroySingleton()
{
	VALIDATE( _isConstructed );
	VALIDATE( !_isInitialized );
	VALIDATE( _TheNetwork );

	/// Destroy the singleton and null the static pointer
	delete _TheNetwork;
	_TheNetwork = NULL;
}


//-------------------------------------------------------------------------------------------------
// Init
//-------------------------------------------------------------------------------------------------
void TheNetwork::Init()
{
	CONSOLE->Printf( "::: Initialization : Network Subsystem (starting)\n" );

	/// Make sure the singleton object doesn't get initialized twice
	VALIDATE( this && _isConstructed && !_isInitialized );

	/// Clear network statics
	memset( &_winSockVersionData, 0, sizeof( _winSockVersionData ) );

	/// Initialize critical sections
	InitializeCriticalSection( &_criticalSection_connectionList );

	/// Set initialization complete
	_isInitialized = true;
	CONSOLE->Printf( "::: Initialization : Network Subsystem (finished)\n" );
}


//-------------------------------------------------------------------------------------------------
// Shutdown
//-------------------------------------------------------------------------------------------------
void TheNetwork::Shutdown()
{
	CONSOLE->Printf( "::: Shutdown : Network Subsystem (starting)\n" );

	/// Delete critical sections
	DeleteCriticalSection( &_criticalSection_connectionList );

	/// De-initialize the system
	_isInitialized = false;
	CONSOLE->Printf( "::: Shutdown : Network Subsystem (finished)\n" );
}


//-------------------------------------------------------------------------------------------------
// InitializeWinSock
//-------------------------------------------------------------------------------------------------
void TheNetwork::_InitializeWinSock()
{
	/// Initialize WinSock
	unsigned short winSockVersion = 0x0101;
	int errorCode = WSAStartup( winSockVersion, &_winSockVersionData );
	switch( errorCode )
	{
		case 0:
			// no error
			break;
		case WSASYSNOTREADY:
			PANIC( "Network Error: WinSock failed to initialize; Windows socket networking not ready! (WSASYSNOTREADY)" );
			break;
		case WSAVERNOTSUPPORTED:
			PANIC_MSG1( "Network Error: WinSock failed to initialize; WinSock version 0x%04x not available! (WSAVERNOTSUPPORTED)", winSockVersion );
			break;
		case WSAEINPROGRESS:
			PANIC( "Network Error: WinSock failed to initialize; A blocking WinSock operation is in progress! (WSAEINPROGRESS)" );
			break;
		case WSAEPROCLIM:
			PANIC( "Network Error: WinSock failed to initialize; Reached the limit of tasks supported by WinSock! (WSAEPROCLIM)" );
			break;
		case WSAEFAULT:
			PANIC( "Network Error: WinSock failed to initialize; Internal error: _winSockVersionData had invalid address! (WSAEFAULT)" );
			break;
		default:
			PANIC_MSG1( "Network Error: WinSock failed to initialize; UNKNOWN ERROR TYPE (#%d)!", errorCode );
			break;
	}
}


//-------------------------------------------------------------------------------------------------
// ShutDownWinSock
//-------------------------------------------------------------------------------------------------
void TheNetwork::_ShutDownWinSock()
{
	/// Shut down WinSock
	int errorCode = WSACleanup();
	switch( errorCode )
	{
		case 0:
			// no error
			break;
		case WSANOTINITIALISED:
			PANIC( "Network Error: WinSock failed to shut down; WSACleanup() called without prior corresponding successful call to WSAStartup()! (WSANOTINITIALISED)" );
			break;
		case WSAENETDOWN:
			PANIC( "Network Error: WinSock failed to shut down; Windows socket networking not ready! (WSAENETDOWN)" );
			break;
		case WSAEINPROGRESS:
			PANIC( "Network Error: WinSock failed to shut down; A blocking WinSock operation is in progress, or the service provider is still processing a callback function! (WSAEINPROGRESS)" );
			break;
		default:
			PANIC( "Network Error: WinSock failed to shut down; UNKNOWN ERROR TYPE!" );
			break;
	}
}


//-------------------------------------------------------------------------------------------------
// StartNetworkServices
//-------------------------------------------------------------------------------------------------
void TheNetwork::StartNetworkServices()
{
	PANIC( "Network Error: We're sorry, the network subsystem has yet to be completed." ); // FIXME: TEMP!!!

	/// Make sure the network isn't already going
	VALIDATE_MSG0( !_isNetworkServiceStarted, "Network Error: StartNetworkServices() called while network services are already running!" );

	/// Initialize WinSock
	_InitializeWinSock();
	_CreateListenSocket();

	_isNetworkServiceStarted = true;
}


//-------------------------------------------------------------------------------------------------
// StopNetworkServices
//-------------------------------------------------------------------------------------------------
void TheNetwork::StopNetworkServices()
{
	VALIDATE_MSG0( _isNetworkServiceStarted, "Network Error: StopNetworkServices() called without a corresponding prior call to StartNetworkServices()!" );

	/// Shut down WinSock
	_ShutDownWinSock();

	_isNetworkServiceStarted = false;
}


//-------------------------------------------------------------------------------------------------
// CreateListenSocket
//-------------------------------------------------------------------------------------------------
void TheNetwork::_CreateListenSocket()
{
	/// Get local computer name
	char hostName[ MAX_HOSTNAME_LEN ];
	gethostname( hostName, sizeof( hostName ) );
	_SetLocalHostName( hostName );

	/// Get local computer info
	_localHostInfo = gethostbyname( _GetLocalHostName().c_str() );

	/// Fill in local socket address info
	memset( &_localSockAddr, 0, sizeof( _localSockAddr ) );
	_localSockAddr.sin_family = _localHostInfo->h_addrtype;
	_localSockAddr.sin_port = htons( _portNumber );

	/// Create a local "listen" socket
	_listenSocket = socket( AF_INET, SOCK_STREAM, 0 );

	/// Bind this socket to the local host (?)
	bind( _listenSocket, (struct sockaddr*) &_localSockAddr, sizeof( _localSockAddr ) );

	/// Designate the maximum number of incoming socket requests to accept
	listen( _listenSocket, _maxNumberOfConnections );

	/// Spawn a dedicated "listen" thread to handle incoming connection requests
	_StartListenThread();
}


//-------------------------------------------------------------------------------------------------
// HandleSocketSelectError
//-------------------------------------------------------------------------------------------------
void TheNetwork::_HandleSocketSelectError( int errorCode )
{
	switch( errorCode )
	{
		case WSANOTINITIALISED:
			PANIC( "Network Error: select(): WinSock not initialized - has NETWORK->StartNetworkServices() been called? (WSANOTINITIALISED)" );
			break;
		case WSAEFAULT:
			PANIC( "Network Error: select(): Bad parms or internal WinSock resource shortfall! (WSAEFAULT)" );
			break;
		case WSAENETDOWN:
			PANIC( "Network Error: select(): The WinSock network subsystem has failed! (WSAENETDOWN)" );
			break;
		case WSAEINVAL:
			PANIC( "Network Error: select(): Invalid time-out specified OR all parms were NULL! (WSAEINVAL)" );
			break;
		case WSAEINTR:
			PANIC( "Network Error: select(): A blocking WinSock call was canceled through WSACancelBlockingCall()! (WSAEINTR)" );
			break;
		case WSAEINPROGRESS:
			PANIC( "Network Error: select(): A blocking WinSock call is in progress, or the service provider is still processing a callback function! (WSAEINPROGRESS)" );
			break;
		case WSAENOTSOCK:
			PANIC( "Network Error: select(): One of the descriptor sets contains an entry that is not a socket! (WSAENOTSOCK)" );
			break;
		default:
			PANIC_MSG1( "Network Error: select(): UNKNOWN ERROR! (#%d)", errorCode );
			break;
	}
}


//-------------------------------------------------------------------------------------------------
// HandleSocketAcceptError
//-------------------------------------------------------------------------------------------------
void TheNetwork::_HandleSocketAcceptError( int errorCode )
{
	switch( errorCode )
	{
		case WSANOTINITIALISED:
			PANIC( "Network Error: accept(): WinSock not initialized - has NETWORK->StartNetworkServices() been called? (WSANOTINITIALISED)" );
			break;
		case WSAENETDOWN:
			PANIC( "Network Error: accept(): The WinSock network subsystem has failed! (WSAENETDOWN)" );
			break;
		case WSAEFAULT:
			PANIC( "Network Error: accept(): Bad parms! (WSAEFAULT)" );
			break;
		case WSAEINTR:
			PANIC( "Network Error: accept(): A blocking WinSock call was canceled through WSACancelBlockingCall()! (WSAEINTR)" );
			break;
		case WSAEINPROGRESS:
			PANIC( "Network Error: accept(): A blocking WinSock call is in progress, or the service provider is still processing a callback function! (WSAEINPROGRESS)" );
			break;
		case WSAEINVAL:
			PANIC( "Network Error: accept(): listen() was not properly invoked before accept()! (WSAEINVAL)" );
			break;
		case WSAEMFILE:
			PANIC( "Network Error: accept(): The queue is nonempty upon entry to accept and there are no descriptors available! (WSAEMFILE)" );
			break;
		case WSAENOBUFS:
			PANIC( "Network Error: accept(): No buffer space is available! (WSAENOBUFS)" );
			break;
		case WSAENOTSOCK:
			PANIC( "Network Error: accept(): The descriptor is not a socket! (WSAENOTSOCK)" );
			break;
		case WSAEOPNOTSUPP:
			PANIC( "Network Error: accept(): The referenced socket is not a type that supports connection-oriented service! (WSAEOPNOTSUPP)" );
			break;
		case WSAEWOULDBLOCK:
			PANIC( "Network Error: accept(): The socket is marked as nonblocking and no connections are present to be accepted! (WSAEWOULDBLOCK)" );
			break;
		default:
			PANIC_MSG1( "Network Error: accept(): UNKNOWN ERROR! (#%d)", errorCode );
			break;
	}
}


//-------------------------------------------------------------------------------------------------
// GetNewUniqueLocalConnectionID
//-------------------------------------------------------------------------------------------------
unsigned int TheNetwork::_GetNewUniqueLocalConnectionID()
{
	unsigned int newID = _nextConnectionID;
	_nextConnectionID ++;
	return newID;
}


//-------------------------------------------------------------------------------------------------
// HandleNewSocket
// 
// NOTE: newHostInfo is passed by value; it is a copy of a local going out of scope elsewhere.
//-------------------------------------------------------------------------------------------------
void TheNetwork::_HandleNewSocket( SOCKET newSocket, struct sockaddr newHostInfo )
{
	/// Enter the thread-safe critical section
	EnterCriticalSection( &_criticalSection_connectionList );

	/// Make sure we don't already have a connection matching this one
	for( unsigned int i = 0; i < _connectionList.size(); i ++ )
	{
		HostConnection& connection = _connectionList[ i ];
		VALIDATE_MSG0( connection.sock != newSocket, "Network Error: New socket is identical to existing socket!" );
		if( connection.hostInfo.sa_family == newHostInfo.sa_family &&
			connection.hostInfo.sa_data[0] == newHostInfo.sa_data[0] &&
			connection.hostInfo.sa_data[1] == newHostInfo.sa_data[1] &&
			connection.hostInfo.sa_data[2] == newHostInfo.sa_data[2] &&
			connection.hostInfo.sa_data[3] == newHostInfo.sa_data[3] &&
			connection.hostInfo.sa_data[4] == newHostInfo.sa_data[4] &&
			connection.hostInfo.sa_data[5] == newHostInfo.sa_data[5] )
		{
			PANIC_MSG4( "Network Error: New host info for %d.%d.%d.%d is identical to host info already in connection list!", newHostInfo.sa_data[2], newHostInfo.sa_data[3], newHostInfo.sa_data[4], newHostInfo.sa_data[5] );
		}
	}

	/// Add this new connection to the end of the list
	HostConnection newConnection;
	newConnection.sock = newSocket;
	newConnection.hostInfo = newHostInfo;
	newConnection.localConnectionID = _GetNewUniqueLocalConnectionID();
	_connectionList.push_back( newConnection );

	/// Launch a new thread dedicated to receiving data over this new connection
	_StartRecvThread( newConnection.localConnectionID );

	/// Leave the thread-safe critical section
	LeaveCriticalSection( &_criticalSection_connectionList );
}


//-------------------------------------------------------------------------------------------------
// ListenThreadDoWork
// 
// This function and threading is necessary because accept() blocks the thread from which
//	it is called, hanging until an incoming connection request is detected.  This thread
//	is therefore dedicated to the simple task of monitoring incoming connections.
//-------------------------------------------------------------------------------------------------
void TheNetwork::_ListenThreadDoWork( void* data )
{
	for( ;; )
	{
		Sleep( 0 ); // play nice with others

		/// Set up some stuff for a call to select()
		struct fd_set socketList;
		struct timeval timeOut;
		socketList.fd_count = 1; // number of sockets to monitor
		socketList.fd_array[ 0 ] = _listenSocket; // monitor this socket
		timeOut.tv_sec = 0; // number of seconds to query before timeout
		timeOut.tv_usec = MAX_MS_THREAD_LOCK * 1000; // number of microseconds to query before timeout

		/// Call select() to check if there are any incoming connections available
		int numSocketsReadable = select( 0, &socketList, NULL, NULL, &timeOut );
		if( numSocketsReadable == 0 )
		{
			/// select() timed out; this is normal (no incoming connections waiting)
			continue;
		}
		else if( numSocketsReadable == SOCKET_ERROR )
		{
			/// Get the error code from WinSock and handle it
			int errorCode = WSAGetLastError();
			_HandleSocketSelectError( errorCode );
			continue;
		}
		else if( numSocketsReadable != 1 )
		{
			PANIC_MSG1( "Network Error: select() returned %d sockets readable, only asked for 1!", numSocketsReadable );
		}

		/// Answer the incoming connection request and establish a new socket
		struct sockaddr newHostInfo;
		int sizeOfNewHostInfo = sizeof( newHostInfo );
		SOCKET newSocket = accept( _listenSocket, &newHostInfo, &sizeOfNewHostInfo );
		if( newSocket == INVALID_SOCKET )
		{
			/// Get the error code from WinSock and handle it
			int errorCode = WSAGetLastError();
			_HandleSocketAcceptError( errorCode );
			continue;
		}

		/// Take action with this new socket
		_HandleNewSocket( newSocket, newHostInfo );
	}
}


//-------------------------------------------------------------------------------------------------
// _RecvThreadDoWork
//-------------------------------------------------------------------------------------------------
void TheNetwork::_RecvThreadDoWork( void* data )
{
	PANIC( "_RecvThreadDoWork : this networking function is not yet implemented!\n" );
/*
	/// Recover disguised connection ID from <data>
	unsigned int connectionID = (unsigned int) data;

	for( ;; )
	{
		Sleep( 0 ); // play nice with others

		/// Enter the thread-safe critical section
		EnterCriticalSection( &_criticalSection_connectionList );

		/// Find the matching entry in the connection list
		unsigned int i;
		for( i = 0; i < _connectionList.size(); i ++ )
		{
			if( _connectionList[ i ].localConnectionID == connectionID )
			{
				/// Found our connection in the connection list
				break;
			}
		}

		/// Make sure we found our connection in the connection list
		if( i >= _connectionList.size() )
		{
			/// Ack!  Couldn't find it!
			PANIC_MSG1( "Network Error: (FIXME) : _RecvThreadDoWork #%d couldn't find itself in connection list!", connectionID );
		}

		/// Get a reference to this connection for temporary use
		HostConnection& connection = _connectionList[ i ];

		/// Set up some stuff for a call to select()
		struct fd_set socketList;
		struct timeval timeOut;
		socketList.fd_count = 1; // number of sockets to monitor
		socketList.fd_array[ 0 ] = connection.sock; // monitor this socket
		timeOut.tv_sec = 0; // number of seconds to query before timeout
		timeOut.tv_usec = MAX_MS_THREAD_LOCK * 1000; // number of microseconds to query before timeout

		/// Call select() to check if there is any incoming data on this pipe waiting to be received
		int numSocketsReadable = select( 0, &socketList, NULL, NULL, &timeOut );
		if( numSocketsReadable == 0 )
		{
			/// select() timed out; this is normal (no incoming data waiting)
			continue;
		}
		else if( numSocketsReadable == SOCKET_ERROR )
		{
			/// Get the error code from WinSock and handle it
			int errorCode = WSAGetLastError();
			_HandleSocketSelectError( errorCode );
			continue;
		}
		else if( numSocketsReadable != 1 )
		{
			PANIC_MSG1( "Network Error: select() returned %d sockets readable, only asked for 1!", numSocketsReadable );
		}
asdf
		/// Receive incoming data into ....?
asdf

WSAAsyncSelect

asdf
		/// Answer the incoming connection request and establish a new socket
		struct sockaddr newHostInfo;
		int sizeOfNewHostInfo = sizeof( newHostInfo );
		SOCKET newSocket = accept( _listenSocket, &newHostInfo, &sizeOfNewHostInfo );
		if( newSocket == INVALID_SOCKET )
		{
			/// Get the error code from WinSock and handle it
			int errorCode = WSAGetLastError();
			_HandleSocketAcceptError( errorCode );
			continue;
		}

		/// Take action with this new socket
		_HandleNewSocket( newSocket, newHostInfo );
	}
*/
}


//-------------------------------------------------------------------------------------------------
// ListenThreadEntryPoint
// 
// This function serves as the entry point for launching the "listen" thread; unfortunately,
//	the C calling convention of _beginthread requires that this function be static (no *this).
// 
//	ListenThreadEntryPoint() therefore does nothing more than call ListenThreadDoWork()
//	on TheNetwork singleton.
//-------------------------------------------------------------------------------------------------
STATIC void TheNetwork::_ListenThreadEntryPoint( void* data )
{
	NETWORK->_ListenThreadDoWork( data );
}


//-------------------------------------------------------------------------------------------------
// _RecvThreadEntryPoint
//-------------------------------------------------------------------------------------------------
STATIC void TheNetwork::_RecvThreadEntryPoint( void* data )
{
	NETWORK->_RecvThreadDoWork( data );
}


//-------------------------------------------------------------------------------------------------
// StartListenThread
//-------------------------------------------------------------------------------------------------
void TheNetwork::_StartListenThread()
{
	_beginthread( _ListenThreadEntryPoint, 0, NULL );
}


//-------------------------------------------------------------------------------------------------
// _StartRecvThread
//-------------------------------------------------------------------------------------------------
void TheNetwork::_StartRecvThread( unsigned int connectionID )
{
	_beginthread( _RecvThreadEntryPoint, 0, (void*) connectionID );
}

