// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.

#include "pch.h"
#include "Base.h"
#include "Image.h"
#include "gameCommon.h"
#include "TheGame.h"
using namespace std;


//-------------------------------------------------------------------------------------------------
// Static member initialization
// 
// (others are also defined in Base_Draw.cpp)
//-------------------------------------------------------------------------------------------------



//-------------------------------------------------------------------------------------------------
// Globals
//-------------------------------------------------------------------------------------------------
list<Base*> baseList;



//-------------------------------------------------------------------------------------------------
// Constructor
//-------------------------------------------------------------------------------------------------
Base::Base()
{
	/// Do we really need this?
}


//-------------------------------------------------------------------------------------------------
// Constructor( int, int )
//-------------------------------------------------------------------------------------------------
Base::Base( const Vector2& position, int gridWidth, int gridHeight )
:
m_width( gridWidth ),
m_height( gridHeight ),
m_pos( position ),
m_isAlive( true ),
m_isDyingFromHeart( false )
{
	m_grid.resize( m_height );
	for( int h = 0; h < m_height; ++ h )
	{
		m_grid[ h ].resize( m_width );
	}

	InitBounds();
	InitNodes();
}



//-------------------------------------------------------------------------------------------------
// Destructor
//-------------------------------------------------------------------------------------------------
Base::~Base()
{
}


//-------------------------------------------------------------------------------------------------
// InitBounds
//-------------------------------------------------------------------------------------------------
void Base::InitBounds()
{
	/// Erect an "empty" (inverted) bounding box initially
	m_bounds.SetFromMinXYMaxXY( FLT_MAX, FLT_MAX, -FLT_MAX, -FLT_MAX );
	m_outerRadius = 0.0f;

	/// Generate a "padding" box for node radius
	Vector2 nodeRadiusBounds( NODE_RADIUS, NODE_RADIUS );

	/// Calc bounds for each node
	for( int x = 0; x < m_width; ++ x )
	{
		for( int y = 0; y < m_height; ++ y )
		{
			BaseNode& node = m_grid[ y ][ x ];
			Vector2 pointNode = m_pos + CalcOffsetForGridXY( x, y );
			node.m_bounds.SetFromPointPadding( pointNode, nodeRadiusBounds );
		}
	}
}


//-------------------------------------------------------------------------------------------------
// InitNodes
//-------------------------------------------------------------------------------------------------
void Base::InitNodes()
{
	for( int x = 0; x < m_width; ++ x )
	{
		for( int y = 0; y < m_height; ++ y )
		{
			BaseNode& node = m_grid[ y ][ x ];
			node.InitVacantNode( this, x, y );
		}
	}
}


//-------------------------------------------------------------------------------------------------
// AddHeart
//-------------------------------------------------------------------------------------------------
void Base::AddHeart( int heart_x, int heart_y )
{
	ASSERT( heart_x >= 0 && heart_x < m_width );
	ASSERT( heart_y >= 0 && heart_y < m_height );

	BaseNode& node = m_grid[ heart_y ][ heart_x ];
	node.m_type = NODETYPE_HEART;
	node.m_hitPoints = HEART_HP;

	m_heartX = heart_x;
	m_heartY = heart_y;

	node.m_isConnectedToHeart = true;

	/// Stretch the overall base bounds by the bounds of this node
	m_bounds.StretchBoundsToIncludeBox( node.m_bounds );
	m_outerRadius = 0.5f * m_bounds.CalcDiagonalLength();
}


//-------------------------------------------------------------------------------------------------
// AddSpawner
//-------------------------------------------------------------------------------------------------
void Base::AddSpawner( int spawner_x, int spawner_y, SpawnType spawnType, double spawnInterval, int maxSpawnCount )
{
	ASSERT( spawner_x >= 0 && spawner_x < m_width );
	ASSERT( spawner_y >= 0 && spawner_y < m_height );

	BaseNode& node = m_grid[ spawner_y ][ spawner_x ];

	if( node.m_spawner )
		return; // don't spawn a second one on the same node

	if( node.m_type == NODETYPE_OUTOFBOUNDS || node.m_type == NODETYPE_EMPTY )
		return; // don't put a spawner on an empty node

	node.m_spawner = new Spawner( spawnType, spawnInterval );
}


//-------------------------------------------------------------------------------------------------
// AddConnection
//-------------------------------------------------------------------------------------------------
void Base::AddConnection( int startX, int startY, int endX, int endY )
{
	ASSERT( startX >= 0 && startX < m_width );
	ASSERT( startY >= 0 && startY < m_height );
	ASSERT( endX >= 0 && endX < m_width );
	ASSERT( endY >= 0 && endY < m_height );

	/// Get convenience references to the nodes at either end
	BaseNode& node1 = m_grid[ startY ][ startX ];
	BaseNode& node2 = m_grid[ endY ][ endX ];

	/// Calculate which direction this connection goes (from both ends)
	HexDirection dirForward = GetDirectionFromXY( endX - startX, endY - startY );
	HexDirection dirBackward = ReverseDirection( dirForward );

	/// Make sure these connections are empty
	ASSERT( node1.m_connections[ dirForward ] == NULL );
	ASSERT( node2.m_connections[ dirBackward ] == NULL );

	/// Upgrade "empty" to "terminator", "terminator" to "junction", etc.
	node1.UpgradeConnectionType();
	node2.UpgradeConnectionType();
	node1.m_numberOfConnections ++;
	node2.m_numberOfConnections ++;

	/// Locate start and end points in world space
	Vector2 startPoint = m_pos + CalcOffsetForGridXY( startX, startY );
	Vector2 endPoint = m_pos + CalcOffsetForGridXY( endX, endY );

	/// Calculate a pipe "radius" padding box
	Vector2 radiusBounds( PIPE_RADIUS, PIPE_RADIUS );

	/// Allocate a new connector
	BasePipe* pipe = new BasePipe();
	pipe->m_gridStartX = startX;
	pipe->m_gridStartY = startY;
	pipe->m_gridEndX = endX;
	pipe->m_gridEndY = endY;
	pipe->m_direction = dirForward;
	pipe->m_bounds.SetFromTwoPointsPadding( startPoint, endPoint, radiusBounds );
	pipe->m_animStartTime = g_timeNow;
	pipe->m_status = PIPESTATUS_INTACT;
	pipe->m_startNode = &node1;
	pipe->m_endNode = &node2;
	pipe->m_base = this;

	/// Assign this connector to both of its endpoint nodes
	node1.m_connections[ dirForward ] = pipe;
	node2.m_connections[ dirBackward ] = pipe;

	/// Stretch overall base bounds by both endpoints' bounding boxes
	m_bounds.StretchBoundsToIncludeBox( node1.m_bounds );
	m_bounds.StretchBoundsToIncludeBox( node2.m_bounds );
	m_outerRadius = 0.5f * m_bounds.CalcDiagonalLength();

	/// Exchange heart-connectedness
	if( node2.m_isConnectedToHeart )
	{
		node1.m_isConnectedToHeart = true;
	}
	if( node1.m_isConnectedToHeart )
	{
		node2.m_isConnectedToHeart = true;
	}
}


//-------------------------------------------------------------------------------------------------
// SpawnBase
//-------------------------------------------------------------------------------------------------
STATIC Base& Base::SpawnBase( const Vector2& position, int gridWidth, int gridHeight )
{
	/// Create a new base
	Base* spawn = new Base( position, gridWidth, gridHeight );
	
	/// Add it to the end of the appropriate list
	baseList.push_back( spawn );

	return *spawn;
}


//-------------------------------------------------------------------------------------------------
// ClearBaseList
//-------------------------------------------------------------------------------------------------
STATIC void Base::ClearBaseList()
{
	while( !baseList.empty() )
	{
		Base* dead = baseList.front();
		baseList.pop_front();
		delete dead;
	}
}


//-------------------------------------------------------------------------------------------------
// Update
//-------------------------------------------------------------------------------------------------
void Base::Update()
{
	/// For each node in this base...
	for( int x = 0; x < m_width; ++ x )
	{
		for( int y = 0; y < m_height; ++ y )
		{

			BaseNode& node = m_grid[ y ][ x ];
			UpdateConnectionsForNode( node );

			/// Check for node death
			if( node.m_hitPoints <= 0 )
			{
				ResolveNodeDeath( node );
				continue;
			}

			/// Update spawner at this node if one is present
			if( node.m_spawner )
			{
				node.UpdateSpawner();
			}

			/// Update the death-shockwave if we're dying
			UpdateShockwave();
		}
	}
}


//-------------------------------------------------------------------------------------------------
// UpdateShockwave
//-------------------------------------------------------------------------------------------------
void Base::UpdateShockwave()
{
	if( !m_isDyingFromHeart )
		return;

	float fractionDead = CalcAnimFraction();
	if( fractionDead <= 0.0f || fractionDead >= 1.0f )
		return;

	float prevFractionDead = CalcPrevAnimFraction();
	if( prevFractionDead < 0.0f || prevFractionDead >= 1.0f )
		return;

	/// For each node in this base...
	for( int x = 0; x < m_width; ++ x )
	{
		for( int y = 0; y < m_height; ++ y )
		{
			BaseNode& node = m_grid[ y ][ x ];
			CheckNodeVersesShockwave( node, prevFractionDead, fractionDead );
		}
	}
}


//-------------------------------------------------------------------------------------------------
// CheckNodeVersesShockwave
//-------------------------------------------------------------------------------------------------
void Base::CheckNodeVersesShockwave( BaseNode& node, float minFractionDead, float maxFractionDead )
{
	float minRadius = minFractionDead * m_outerRadius;
	float maxRadius = maxFractionDead * m_outerRadius;

	float minSquared = (minRadius * minRadius);
	float maxSquared = (maxRadius * maxRadius);

	/// Get squared-distance from node center to base center
	float distSquared = DistanceSquaredBetweenPoints2d( m_bounds.CalcCenter(), node.m_pos );

	/// Ignore if distSquared is outside the range [minSquared, maxSquared]
	if( (distSquared < minSquared) || (distSquared > maxSquared) )
		return;

	/// Reduce this node to 0 hp
	node.m_hitPoints = 0;
}


//-------------------------------------------------------------------------------------------------
// UpdateConnectionForNodeDirection
//-------------------------------------------------------------------------------------------------
void Base::UpdateConnectionForNodeDirection( BaseNode& node, HexDirection direction )
{
	BasePipe* pipe;
	
	/// Bugger out if there's no connection here
	pipe = node.m_connections[ direction ];
	if( !pipe )
	{
		return;
	}

	/// Update the pipe, and kill it if requested to do so
	bool killPipe = UpdatePipe( *pipe );
	if( killPipe )
	{
		/// Destroy the pipe; this should be safe, since UpdatePipe returns true
		//	if and only if the pipe has been detached / scheduled for destruction.
		delete pipe;
	}
}


//-------------------------------------------------------------------------------------------------
// UpdateConnectionsForNode
//-------------------------------------------------------------------------------------------------
void Base::UpdateConnectionsForNode( BaseNode& node )
{
	UpdateConnectionForNodeDirection( node, HEXDIRECTION_EAST );
	UpdateConnectionForNodeDirection( node, HEXDIRECTION_NORTHEAST );
	UpdateConnectionForNodeDirection( node, HEXDIRECTION_NORTHWEST );
}


//-------------------------------------------------------------------------------------------------
// UpdatePipe
// 
// Return true if pipe should be destroyed, false if not.
//-------------------------------------------------------------------------------------------------
bool Base::UpdatePipe( BasePipe& pipe )
{
	ASSERT( pipe.m_startNode && pipe.m_endNode );

	/// Check if this pipe is dying
	if( pipe.IsDying() )
	{
		/// See if the pipe has finished dying yet
		float fraction = pipe.CalcAnimFraction();
		if( fraction >= 1.0f )
		{
			/// Determine which node the pipe was burning down towards
			BaseNode* burnFromNode = pipe.m_startNode;
			BaseNode* burnToNode = pipe.m_endNode;
			if( pipe.m_status == PIPESTATUS_DYING_BACKWARD )
			{
				burnFromNode = pipe.m_endNode;
				burnToNode = pipe.m_startNode;
			}

			HexDirection dirFromFromNode = burnFromNode->FindPipeAmongConnections( &pipe );
			HexDirection dirFromToNode = burnToNode->FindPipeAmongConnections( &pipe );
			ASSERT( dirFromFromNode != HEXDIRECTION_NONE && dirFromToNode != HEXDIRECTION_NONE );

			/// Remove references to this pipe from both nodes
			burnFromNode->UnhookPipe( &pipe );
			burnToNode->UnhookPipe( &pipe );

			ContinueChainReactionAtNode( *burnToNode );
			return true;
		}
	}

	return false;
}


//-------------------------------------------------------------------------------------------------
// ContinueChainReactionAtNode
//-------------------------------------------------------------------------------------------------
void Base::ContinueChainReactionAtNode( BaseNode& node )
{
	/// Stop burning if this is a junction (unless it's a full-base heart-burn)
	if( node.m_numberOfConnections > 1 && !m_isDyingFromHeart )
		return;

	/// Don't allow a chain-reaction to damage a heart node (unless it's the last piece)
	if( node.m_numberOfConnections > 0 && node.m_type == NODETYPE_HEART )
		return;

	for( int i = 0; i < NUM_HEXDIRECTIONS; ++ i )
	{
		if( node.m_connections[ i ] )
			continue;

		node.m_hitPoints = 0;
		return;
	}
}




//-------------------------------------------------------------------------------------------------
// ResolveNodeDeath
//-------------------------------------------------------------------------------------------------
void Base::ResolveNodeDeath( BaseNode& node )
{
	/// Only process death once for each node
	if( node.IsEmpty() )
		return;

	/// Check if this was the heart node of the base
	if( node.m_type == NODETYPE_HEART )
	{
		OnHeartDeath();
	}

	node.m_type = NODETYPE_EMPTY;
	node.m_hitPoints = 0;

	/// Start death anims for all connectors attached to this node
	for( int dir = 0; dir < NUM_HEXDIRECTIONS; ++ dir )
	{
		BasePipe* pipe = node.m_connections[ dir ];

		/// Skip if no connection exists in that direction
		if( !pipe )
			continue;

		/// Skip if this pipe is already dying
		if( pipe->IsDying() )
			continue;

		/// Determine if the dying node is at the front or back of the pipe
		ASSERT( pipe->m_startNode == &node || pipe->m_endNode == &node );
		if( pipe->m_startNode == &node )
		{
			pipe->m_status = PIPESTATUS_DYING_FORWARD;
		}
		else
		{
			pipe->m_status = PIPESTATUS_DYING_BACKWARD;
		}

		pipe->m_animStartTime = g_timeNow;
	}
}


//-------------------------------------------------------------------------------------------------
// OnHeartDeath
//-------------------------------------------------------------------------------------------------
void Base::OnHeartDeath()
{
	m_isDyingFromHeart = true;
	m_timeOfDeath = g_timeNow;
}


//-------------------------------------------------------------------------------------------------
// ReserveGridNode
//-------------------------------------------------------------------------------------------------
void Base::ReserveGridNode( int x, int y )
{
	m_grid[ y ][ x ].m_type = NODETYPE_EMPTY;
}


//-------------------------------------------------------------------------------------------------
// Generate
// 
// Use the base's current growth heuristics to grow an entire base.  If heuristics
//	other than the defaults are desired, a BaseGrowthHeuristic should first be
//	created prior to calling Generate() and passed in as a parameter.
//-------------------------------------------------------------------------------------------------
void Base::Generate( const BaseGrowthHeuristic& heuristic )
{
	GenerateUsingTestHeuristic( heuristic );
}


//-------------------------------------------------------------------------------------------------
// GenerateUsingTestHeuristic
//-------------------------------------------------------------------------------------------------
void Base::GenerateUsingTestHeuristic( const BaseGrowthHeuristic& heuristic )
{
	/// Calculate the gridwise "radius" of this base
	int centerX = m_width / 2;
	int centerY = m_height / 2;
	int radius = Max( centerX, centerY );
	list<int> sorted;
	list<int> scrambled;

	/// Create a heart at the base's center
	AddHeart( centerX, centerY );

	/// Mark out the general shape of the base with "empty" nodes
	for( int y = 0; y < m_height; y ++ )
	{
		for( int x = 0; x < m_width; x ++ )
		{
			/// Continue if this is the heart
			if( (x == m_heartX) && (y == m_heartY) )
				continue;

			int distance = CalcGridDistBetweenPoints( x, y, centerX, centerY );
			if( distance > radius )
				continue;

			m_grid[ y ][ x ].m_type = NODETYPE_EMPTY;
			sorted.push_back( y * m_width + x ); // build a list of valid grid nodes
		}
	}

	/// Scramble the order of the nodes
	list<int>::iterator iter;
	while( !sorted.empty() )
	{
		/// Pick a random entry
		int size = sorted.size();
		int randomIndex = rand() % size;
		int removed = 0;

		/// Remove it from the sorted list
		for( iter = sorted.begin(); iter != sorted.end(); ++ iter )
		{
			if( randomIndex == 0 )
			{
				removed = *iter;
				sorted.remove( *iter );
				break;
			}

			randomIndex --;
		}

		scrambled.push_back( removed );
	}

	/// Perform some algorithm on each empty node, in random order, until all have been visited
	while( !scrambled.empty() )
	{
		int id = scrambled.front();
		scrambled.pop_front();
		int x = id % m_width;
		int y = id / m_width;

		GenerateStuffAtNode( x, y, heuristic );
	}

	AddSpawner( m_heartX, m_heartY, OBJTYPE_GRUNT );
/*	/// Place (up to) N random spawners
	for( int i = 0; i < 1; i ++ )
	{
		BaseNode* node = GetRandomOccupiedNode();
		AddSpawner( node->m_gridX, node->m_gridY );
	}
*/
}


//-------------------------------------------------------------------------------------------------
// GetRandomOccupiedNode
//-------------------------------------------------------------------------------------------------
BaseNode* Base::GetRandomOccupiedNode()
{
	int x, y;

	/// Count total number of valid nodes
	int nodeCount = 0;
	for( x = 0; x < m_width; x ++ )
	{
		for( y = 0; y < m_height; y ++ )
		{
			BaseNode& node = m_grid[ y ][ x ];
			if( node.IsEmpty() )
				continue;

			nodeCount ++;
		}
	}

	if( !nodeCount )
		return NULL;

	/// Pick a random node number
	int killNodeNum = rand() % nodeCount;
	for( x = 0; x < m_width; x ++ )
	{
		for( y = 0; y < m_height; y ++ )
		{
			BaseNode& node = m_grid[ y ][ x ];
			if( node.IsEmpty() )
				continue;

			if( killNodeNum <= 0 )
			{
				return &node;
			}

			killNodeNum --;
		}
	}

	return NULL;
}


//-------------------------------------------------------------------------------------------------
// GenerateStuffAtNode
//-------------------------------------------------------------------------------------------------
void Base::GenerateStuffAtNode( int x, int y, const BaseGrowthHeuristic& heuristic )
{
	BaseNode& node = m_grid[ y ][ x ];

	/// Do nothing if this node is already connected to the heart
	if( node.m_isConnectedToHeart )
		return; // STOP GENERATING - WE'RE CONNECTED
//	if( grid[ y ][ x ].type != NODETYPE_EMPTY )
//		return;

	/// Score all surrounding nodes
	float highestScore = 0.0f;
	int highestDirection = HEXDIRECTION_NONE;
	int highestX, highestY;
	for( int dir = 0; dir < NUM_HEXDIRECTIONS; dir ++ )
	{
		int dirX, dirY;
		GetXYFromDirection( (HexDirection) dir, dirX, dirY );
		int adjX = x + dirX;
		int adjY = y + dirY;

		/// Reject if dirX,dirY is off the grid
		if( (adjX < 0) || (adjX >= m_width) || (adjY < 0) || (adjY >= m_height) )
			continue;

		/// Reject if dirX,dirY is marked as out-of-bounds
		BaseNode& adjNode = m_grid[ adjY ][ adjX ];
		if( adjNode.m_type == NODETYPE_OUTOFBOUNDS )
			continue;

		/// Reject if this node already has a connection for that direction
		if( node.m_connections[ dir ] != NULL )
			continue;

		/// Reject if that is a non-empty, but not heart-connected, node (prevents loop-back)
		if( adjNode.m_type != NODETYPE_EMPTY && adjNode.m_isConnectedToHeart == false )
			continue;

		float score = ScoreConnectionDirectionFromXY( (HexDirection) dir, x, y, heuristic );
		if( score > highestScore )
		{
			highestScore = score;
			highestDirection = dir;
			highestX = adjX;
			highestY = adjY;
		}
	}

	/// Do nothing if no directions were able to be scored
	if( highestDirection == HEXDIRECTION_NONE )
		return;

	/// Add connection in that direction
	AddConnection( x, y, highestX, highestY );

	/// Check if we are now connected (directly or indirectly) to the heart
	if( node.m_isConnectedToHeart )
		return;

	/// Recurse from that node; this only returns when heart-connection is established
	GenerateStuffAtNode( highestX, highestY, heuristic );
	node.m_isConnectedToHeart = true;
	return;
}


//-------------------------------------------------------------------------------------------------
// ScoreConnectionDirectionFromXY
//-------------------------------------------------------------------------------------------------
float Base::ScoreConnectionDirectionFromXY( HexDirection dir, int startX, int startY, const BaseGrowthHeuristic& heuristic )
{
	float score = 0.0f;

	int dirX, dirY;
	GetXYFromDirection( dir, dirX, dirY );
	int endX = startX + dirX;
	int endY = startY + dirY;
	BaseNode& startNode = m_grid[ startY ][ startX ];
	BaseNode& endNode = m_grid[ endY ][ endX ];
	BaseNodeType startType = startNode.m_type;
	BaseNodeType endType = endNode.m_type;

	int startDistFromHeart = CalcGridDistBetweenPoints( startX, startY, m_heartX, m_heartY );
	int endDistFromHeart = CalcGridDistBetweenPoints( endX, endY, m_heartX, m_heartY );

	int algorithm = 1;

	//-------------------------------------------------------------------------------------------------
	// (terminator->heart->junction->empty), closer
	//-------------------------------------------------------------------------------------------------
	if( algorithm == 1 )
	{
		/// Score closeness
		if( endDistFromHeart < startDistFromHeart )
		{
			score += 30.0f;
		}
		else if( endDistFromHeart == startDistFromHeart )
		{
			score += 20.0f;
		}
		else
		{
			score += 10.0f;
		}

		/// Score emptiness
		if( endType == NODETYPE_EMPTY )
		{
			score += 0.0f;
		}
		else if( endType == NODETYPE_HEART )
		{
			score += 300.0f;
		}
		else if( endType == NODETYPE_TERMINATOR )
		{
			score += 400.0f;
		}
		else if( endType == NODETYPE_JUNCTION )
		{
			score += 200.0f;
		}
	}

	return score;
}


//-------------------------------------------------------------------------------------------------
// CalcAnimFraction
//-------------------------------------------------------------------------------------------------
float Base::CalcAnimFraction()
{
	float animFraction = 0.0f;
	if( m_isDyingFromHeart )
	{
		double animStart = m_timeOfDeath;
		double animFinish = animStart + BASE_DEATH_SECONDS;
		animFraction = (float)( (g_timeNow - animStart) / (animFinish - animStart) );
		animFraction = ClampFloat( animFraction, 0.0f, 1.0f );
	}

	return animFraction;
}


//-------------------------------------------------------------------------------------------------
// CalcPrevAnimFraction
//-------------------------------------------------------------------------------------------------
float Base::CalcPrevAnimFraction()
{
	float animFraction = 0.0f;
	if( m_isDyingFromHeart )
	{
		double animStart = m_timeOfDeath;
		double animFinish = animStart + BASE_DEATH_SECONDS;
		double timeLast = g_timeNow - g_frameSeconds;
		animFraction = (float)( (timeLast - animStart) / (animFinish - animStart) );
		animFraction = ClampFloat( animFraction, 0.0f, 1.0f );
	}

	return animFraction;
}


