// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.

#include "pch.h"
#include "Base.h"
#include "Image.h"
#include "gameCommon.h"
#include "TheGame.h"
#include "Player.h"


//-------------------------------------------------------------------------------------------------
// Static member initialization
// 
// (others are also defined in Base.cpp)
//-------------------------------------------------------------------------------------------------
STATIC const Rgba Base::COLOR_TERMINATOR_LIGHT_EMPTY( 0.5f, 0.5f, 0.5f, 0.5f );
STATIC const Rgba Base::COLOR_TERMINATOR_DARK_EMPTY( 0.3f, 0.3f, 0.3f, 0.5f );
STATIC const Rgba Base::COLOR_TERMINATOR_LIGHT_JUNCTION( 0.6f, 0.6f, 0.6f, 1.0f );
STATIC const Rgba Base::COLOR_TERMINATOR_DARK_JUNCTION( 0.3f, 0.3f, 0.3f, 1.0f );
STATIC const Rgba Base::COLOR_TERMINATOR_LIGHT_TERMINATOR( 0.3f, 0.3f, 1.0f, 1.0f );
STATIC const Rgba Base::COLOR_TERMINATOR_DARK_TERMINATOR( 0.2f, 0.2f, 0.6f, 1.0f );
STATIC const Rgba Base::COLOR_TERMINATOR_LIGHT_HEART( 1.0f, 0.0f, 1.0f, 1.0f );
STATIC const Rgba Base::COLOR_TERMINATOR_DARK_HEART( 0.5f, 0.0f, 0.5f, 1.0f );
STATIC const Rgba Base::COLOR_CONNECTOR_LIGHT( 0.6f, 0.6f, 0.6f, 1.0f );
STATIC const Rgba Base::COLOR_CONNECTOR_DARK( 0.3f, 0.3f, 0.3f, 1.0f );


//-------------------------------------------------------------------------------------------------
// GetLightColorForNodeType
//-------------------------------------------------------------------------------------------------
Rgba Base::GetLightColorForNodeType( BaseNodeType type )
{
	if( type == NODETYPE_EMPTY || type == NODETYPE_OUTOFBOUNDS )
	{
		return COLOR_TERMINATOR_LIGHT_EMPTY;
	}
	else if( type == NODETYPE_JUNCTION )
	{
		return COLOR_TERMINATOR_LIGHT_JUNCTION;
	}
	else if( type == NODETYPE_TERMINATOR )
	{
		return COLOR_TERMINATOR_LIGHT_TERMINATOR;
	}
	else if( type == NODETYPE_HEART )
	{
		return COLOR_TERMINATOR_LIGHT_HEART;
	}

	PANIC_MSG1( "GetLightColorForNodeType : unknown node type # %d", type );
}


//-------------------------------------------------------------------------------------------------
// GetDarkColorForNodeType
//-------------------------------------------------------------------------------------------------
Rgba Base::GetDarkColorForNodeType( BaseNodeType type )
{
	if( type == NODETYPE_EMPTY || type == NODETYPE_OUTOFBOUNDS )
	{
		return COLOR_TERMINATOR_DARK_EMPTY;
	}
	else if( type == NODETYPE_JUNCTION )
	{
		return COLOR_TERMINATOR_DARK_JUNCTION;
	}
	else if( type == NODETYPE_TERMINATOR )
	{
		return COLOR_TERMINATOR_DARK_TERMINATOR;
	}
	else if( type == NODETYPE_HEART )
	{
		return COLOR_TERMINATOR_DARK_HEART;
	}

	PANIC_MSG1( "GetDarkColorForNodeType : unknown node type # %d", type );
}


//-------------------------------------------------------------------------------------------------
// DrawNode
//-------------------------------------------------------------------------------------------------
void Base::DrawNode( const Vector2& centerPos, BaseNodeType type )
{
	const int numSides = 12;

	if( type == NODETYPE_EMPTY || type == NODETYPE_OUTOFBOUNDS )
//	if( type == NODETYPE_OUTOFBOUNDS )
	{
		return;
	}

	float radius = NODE_RADIUS;
	if( type == NODETYPE_JUNCTION )
	{
		radius = JUNCTION_RADIUS;
	}

	Rgba colorLight = GetLightColorForNodeType( type );
	Rgba colorDark = GetDarkColorForNodeType( type );

	glDisable( GL_TEXTURE_2D );
	glShadeModel( GL_SMOOTH );
	glBegin( GL_TRIANGLE_FAN );
	RENDERER->SetColor( colorLight );
	glVertex2f( centerPos.x, centerPos.y );
	for( int i = 0; i <= numSides; ++ i )
	{
		RENDERER->SetColor( colorDark );
		float f = float( i ) / float( numSides );
		Vector2 offset;
		offset.SetFromCosSinRadians( (float) TWO_PI * f );
		offset *= radius;
		const Vector2 pos = centerPos + offset;
		glVertex2f( pos.x, pos.y );
	}
	glEnd();
}



//-------------------------------------------------------------------------------------------------
// DrawPipe
//-------------------------------------------------------------------------------------------------
void Base::DrawPipe( BasePipe& pipe )
{
	float fractionDead = pipe.CalcAnimFraction();
	Vector2 startPos = pipe.m_startNode->m_pos;
	Vector2 endPos = pipe.m_endNode->m_pos;
	Vector2 explosionPoint;
	bool drawExplosion = false;

	/// Check if the pipe is dying; if so, shore up either <startPos> or <endPos>
	if( pipe.m_status == PIPESTATUS_DYING_FORWARD )
	{
		/// Move <startPos> towards <endPos> based on <fractionDead>
		Vector2 dispFromEnd = startPos - endPos;
		dispFromEnd *= (1.0f - fractionDead);
		startPos = endPos + dispFromEnd;
		explosionPoint = startPos;
		drawExplosion = true;
	}
	else if( pipe.m_status == PIPESTATUS_DYING_BACKWARD )
	{
		/// Move <startPos> towards <endPos> based on <fractionDead>
		Vector2 dispFromStart = endPos - startPos;
		dispFromStart *= (1.0f - fractionDead);
		endPos = startPos + dispFromStart;
		explosionPoint = endPos;
		drawExplosion = true;
	}

	Rgba colorLight = COLOR_CONNECTOR_LIGHT;
	Rgba colorDark = COLOR_CONNECTOR_DARK;
	float alpha = 1.0f;

	colorLight.a = alpha;
	colorDark.a = alpha;

	const Vector2 startToEnd = endPos - startPos;
	Vector2 forward( startToEnd );
	const float len = forward.Normalize();
	forward *= PIPE_RADIUS;
	Vector2 right( forward.y, -forward.x );

	Vector2 point1 = (startPos) - right;
	Vector2 point2 = (startPos);
	Vector2 point3 = (startPos) + right;
	Vector2 point4 = (endPos) + right;
	Vector2 point5 = (endPos);
	Vector2 point6 = (endPos) - right;

	glDisable( GL_TEXTURE_2D );
	glShadeModel( GL_SMOOTH );
	glBegin( GL_QUAD_STRIP );

	RENDERER->SetColor( colorDark );
	glVertex2f( point1.x, point1.y );
	glVertex2f( point6.x, point6.y );

	RENDERER->SetColor( colorLight );
	glVertex2f( point2.x, point2.y );
	glVertex2f( point5.x, point5.y );

	RENDERER->SetColor( colorDark );
	glVertex2f( point3.x, point3.y );
	glVertex2f( point4.x, point4.y );

	glEnd();

	if( drawExplosion )
	{
		DrawExplosionAtPoint( explosionPoint );
	}
}


//-------------------------------------------------------------------------------------------------
// DrawExplosionAtPoint
//-------------------------------------------------------------------------------------------------
void Base::DrawExplosionAtPoint( const Vector2& point )
{
	int numParticles = 30;
	float minRadius = 0.0f;
	float maxRadius = 10.0f;

	if( m_isDyingFromHeart )
	{
		numParticles = 50;
		minRadius = 2.0f;
		maxRadius = 15.0f;
	}

	glDisable( GL_POINT_SMOOTH );
	glPointSize( 6.0f );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE );
	glBegin( GL_POINTS );
	glColor4f( 1.0f, 0.5f, 0.1f, 0.5f );

	for( int i = 0; i < numParticles; i ++ )
	{
		float r = RandomFloat( minRadius, maxRadius );
		float radians = RandomFloat( 0.0f, (float) TWO_PI );
		Vector2 pos;
		pos.SetFromCosSinRadians( radians );
		pos *= r;
		pos += point;

		glVertex2f( pos.x, pos.y );
	}

	glEnd(); // GL_POINTS
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	glPointSize( 1.0f );
}


//-------------------------------------------------------------------------------------------------
// Draw
//-------------------------------------------------------------------------------------------------
void Base::Draw()
{
	int x, y;

//	RENDERER->DrawAABB2( bounds, Rgba::RED );
//	RENDERER->DrawCircle( bounds.CalcCenter(), outerRadius, Rgba::RED );

	Vector2 baseCenter = m_bounds.CalcCenter();
	Vector2 screenCenterToBaseCenter = baseCenter - g_player.m_pos;
	float distanceSquaredToBase = screenCenterToBaseCenter | screenCenterToBaseCenter;
	if( distanceSquaredToBase > (640.0f + m_outerRadius) * (640.0f + m_outerRadius) )
		return;

	/// Draw each connection
	for( x = 0; x < m_width; ++ x )
	{
		for( y = 0; y < m_height; ++ y )
		{
			DrawConnectionsForGridXY( x, y );
		}
	}

	/// Draw each node
	for( x = 0; x < m_width; ++ x )
	{
		for( y = 0; y < m_height; ++ y )
		{
			const BaseNode& node = m_grid[ y ][ x ];
			Vector2 point = m_pos + CalcOffsetForGridXY( x, y );
			DrawNode( point, node.m_type );
		}
	}

	/// Draw death explosion shockwave (if we're dying)
	float fractionDead = CalcAnimFraction();
	if( fractionDead )
	{
		DrawShockwave( fractionDead );
	}
}


//-------------------------------------------------------------------------------------------------
// DrawShockwave
//-------------------------------------------------------------------------------------------------
void Base::DrawShockwave( float fractionDead )
{
	if( fractionDead <= 0.0f || fractionDead >= 1.0f )
		return;

	RENDERER->XYPrintf( m_bounds.CalcCenter().x, m_bounds.CalcCenter().y, "Base is dying (%.0f%%)", 100.0f * fractionDead );

	const int NUM_RINGS = 32;
	const float eachRingFraction = 2.0f / (float) NUM_RINGS;
//	float ringRadii[ NUM_RINGS ];
	for( int i = 0; i < NUM_RINGS; i ++ )
	{
		float ringFraction = fractionDead * (2.0f - (eachRingFraction * (float) i) );
		if( ringFraction <= 0.0f )
			continue;

		float ringRadius = m_outerRadius * ringFraction;

		glLineWidth( 6.0f );
		glBlendFunc( GL_SRC_ALPHA, GL_ONE );
		RENDERER->DrawCircle( m_bounds.CalcCenter(), ringRadius, Rgba( 0.4f, 0.1f, 0.0f, 1.0f - fractionDead ) );
		glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
		glLineWidth( 1.0f );
	}

//	glLineWidth( 10.0f );
//	RENDERER->DrawCircle( m_bounds.CalcCenter(), m_outerRadius * fractionDead, Rgba::RED );
//	glLineWidth( 1.0f );
}


//-------------------------------------------------------------------------------------------------
// DrawConnectionsForGridXY
//-------------------------------------------------------------------------------------------------
void Base::DrawConnectionsForGridXY( int x, int y )
{
	BaseNode& node = m_grid[ y ][ x ];
	BasePipe* pipe;

	/// Draw the connection to the east of this node
	pipe = node.m_connections[ HEXDIRECTION_EAST ];
	if( pipe )
	{
		DrawPipe( *pipe );
	}

	/// Draw the connection to the northeast of this node
	pipe = node.m_connections[ HEXDIRECTION_NORTHEAST ];
	if( pipe )
	{
		DrawPipe( *pipe );
	}

	/// Draw the connection to the northwest of this node
	pipe = node.m_connections[ HEXDIRECTION_NORTHWEST ];
	if( pipe )
	{
		DrawPipe( *pipe );
	}
}


