// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.

#include "pch.h"
#include "Projectiles.h"
#include "gameCommon.h"
#include "GameBalance.h"
#include "BaseCommon.h"
using namespace std;



//-------------------------------------------------------------------------------------------------
// Static members
//-------------------------------------------------------------------------------------------------
list<Projectile*> projectileList[ PROJECTILETYPE_NUM_TYPES ];




//-------------------------------------------------------------------------------------------------
// Globals
//-------------------------------------------------------------------------------------------------
double g_particleTypeLifespanTable[ PROJECTILETYPE_NUM_TYPES ] =
{
	LIFESPAN_PLAYER_BULLETS,
};

float g_particleTypeRadiusTable[ PROJECTILETYPE_NUM_TYPES ] =
{
	RADIUS_PLAYER_BULLETS,
};

float g_particleTypeDamageTable[ PROJECTILETYPE_NUM_TYPES ] =
{
	DAMAGE_PLAYER_BULLETS,
};


//-------------------------------------------------------------------------------------------------
// Projectile
//-------------------------------------------------------------------------------------------------
Projectile::Projectile()
{
}


//-------------------------------------------------------------------------------------------------
// Projectile
//-------------------------------------------------------------------------------------------------
Projectile::Projectile( const Vector2& position, const Vector2& velocity, ProjectileType projectileType, float projectileDensity )
:
m_pos( position ),
m_vel( velocity ),
m_type( projectileType ),
m_density( projectileDensity ),
m_birthDate( g_timeNow ),
m_moveVector( m_vel * (float) g_frameSeconds ),
m_endPos( m_pos + m_moveVector ),
m_moveBounds( m_pos ),
m_moveBoundsFreshnessFrameNumber( APPLICATION->GetFrameNumber() )
{
	float radius = g_particleTypeRadiusTable[ projectileType ];
	m_moveBounds.StretchBoundsToIncludePoint( m_endPos );
	m_moveBounds.AddPadding( radius, radius );
}



//-------------------------------------------------------------------------------------------------
// Copy constructor
//-------------------------------------------------------------------------------------------------
Projectile::Projectile( const Projectile& projectile )
:
m_pos( projectile.m_pos ),
m_vel( projectile.m_vel ),
m_type( projectile.m_type ),
m_density( projectile.m_density ),
m_birthDate( projectile.m_birthDate ),
m_moveVector( projectile.m_moveVector ),
m_endPos( projectile.m_endPos ),
m_moveBounds( projectile.m_moveBounds ),
m_moveBoundsFreshnessFrameNumber( projectile.m_moveBoundsFreshnessFrameNumber )
{
}



//-------------------------------------------------------------------------------------------------
// ~Projectile
//-------------------------------------------------------------------------------------------------
Projectile::~Projectile()
{
}


//-------------------------------------------------------------------------------------------------
// SpawnProjectile
//-------------------------------------------------------------------------------------------------
STATIC const Projectile& Projectile::SpawnProjectile( const Vector2& position, const Vector2& velocity, ProjectileType projectileType, float density )
{
	/// Create a new projectile
	Projectile* spawn = new Projectile( position, velocity, projectileType, density );
	
	/// Add it to the end of the appropriate list
	projectileList[ projectileType ].push_back( spawn );

	return *spawn;
}


//-------------------------------------------------------------------------------------------------
// ExpireParticles
//-------------------------------------------------------------------------------------------------
STATIC void Projectile::ExpireParticles()
{
	double timeNow = g_timeNow;

	/// Loop through each type of projectile
	for( int particleType = 0; particleType < PROJECTILETYPE_NUM_TYPES; ++ particleType )
	{
		/// Get the lifespan for this particle type
		double lifespan = g_particleTypeLifespanTable[ particleType ];
		while( !projectileList[ particleType ].empty() )
		{
			/// Check if the oldest particles (those at the head) are ready to expire
			Projectile* projectile = projectileList[ particleType ].front();
			double age = timeNow - projectile->m_birthDate;
			if( age < lifespan )
			{
				/// No particles are yet old enough to expire
				break;
			}

			projectileList[ particleType ].pop_front();
			delete projectile;
		}
	}
}


//-------------------------------------------------------------------------------------------------
// UpdateParticles
// 
// This is called after collions has already been run for this frame.
// This does the following:
//		- Check if any particles have expired due to age
//		- Move the particle's position (pos) forward by its frametime-adjusted move amount
//		- Adjust / recalculate velocity if necessary (and update moveVector)
//		- Update the new movement endpoint
//		- Update the AABB move bounds
//-------------------------------------------------------------------------------------------------
STATIC void Projectile::UpdateParticles()
{
	/// Check for particle death
	ExpireParticles();

	float frameTime = (float) g_frameSeconds;

	/// For each type of projectile...
	for( int particleType = 0; particleType < PROJECTILETYPE_NUM_TYPES; ++ particleType )
	{
		/// For each live projectile...
		list<Projectile*>::iterator iter;
		for( iter = projectileList[ particleType ].begin(); iter != projectileList[ particleType ].end(); iter ++ )
		{
			Projectile* p = *iter;

			/// Move it forward along its path
			p->m_pos = p->m_endPos;
			
			/// Make adjustments to the velocity vector if needed
			// p->vel += ???, *= ???

			/// Recalculate moveVector based on adjusted vel and frameTime
			p->m_moveVector = p->m_vel * frameTime;

			/// Recalculate new endPos based on new moveVector
			p->m_endPos = p->m_pos + p->m_moveVector;

			/// Don't rebuild the bullet movement bounds; this gets built on demand later
		}
	}
}


//-------------------------------------------------------------------------------------------------
// UpdateMoveBounds
//-------------------------------------------------------------------------------------------------
void Projectile::UpdateMoveBounds()
{
	/// Check if the bounds are already up-to-date
	unsigned int thisFrame = APPLICATION->GetFrameNumber();
	if( m_moveBoundsFreshnessFrameNumber == thisFrame )
		return;

	/// Otherwise, recalculate the bounds and update the "freshness" frame number
	const float radius = g_particleTypeRadiusTable[ m_type ];
	const Vector2 radiusBounds( radius, radius );
	m_moveBounds.SetFromTwoPointsPadding( m_pos, m_endPos, radiusBounds );
	m_moveBoundsFreshnessFrameNumber = thisFrame;
}


