// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.

#include "pch.h"
#include "TheGame.h"
#include "Image.h"
#include "Player.h"
#include "Base.h"
#include "Capsule2.h"
#include "AABB2.h"
#include "Projectiles.h"
#include "LineSegment2.h"
using namespace std;


bool g_debugDrawCollisions = false;


//-------------------------------------------------------------------------------------------------
// ResolveCollisionBulletVsNode
//-------------------------------------------------------------------------------------------------
void TheGame::ResolveCollisionBulletVsNode( Base& base, Projectile& bullet, BaseNode& node )
{
	if( g_debugDrawCollisions )
	{
		glLineWidth( 2.0f );
		RENDERER->DrawAABB2( node.m_bounds );
		RENDERER->DrawAABB2( bullet.m_moveBounds, Rgba::GREY );
		glLineWidth( 1.0f );
	}

	/// Do damage to the node
	int damage = int( g_particleTypeDamageTable[ bullet.m_type ] * bullet.m_density );
	node.m_hitPoints -= damage;
}


//-------------------------------------------------------------------------------------------------
// ResolveCollisionBulletVsPipe
//-------------------------------------------------------------------------------------------------
void TheGame::ResolveCollisionBulletVsPipe( Base& base, Projectile& bullet, BasePipe& pipe, LineSegment2& pipeSegment )
{
	if( g_debugDrawCollisions )
	{
		glLineWidth( 2.0f );
		RENDERER->DrawAABB2( bullet.m_moveBounds, Rgba::GREY );
		RENDERER->DrawVector( pipeSegment );
		glLineWidth( 1.0f );
	}
}


//-------------------------------------------------------------------------------------------------
// DebugDrawCollisionBulletVsBase
//-------------------------------------------------------------------------------------------------
void DebugDrawCollisionBulletVsBase( Base& base, Projectile& bullet )
{
	int x, y;

	/// Loop through each node in the base
	for( x = 0; x < base.m_width; x ++ )
	{
		for( y = 0; y < base.m_height; y ++ )
		{
			BaseNode& node = base.m_grid[ y ][ x ];

			if( (node.m_type == NODETYPE_TERMINATOR || node.m_type == NODETYPE_HEART) && bullet.m_moveBounds.IsOverlapping( node.m_bounds ) )
			{
				RENDERER->DrawAABB2( bullet.m_moveBounds, Rgba::CYAN );
				RENDERER->DrawAABB2( node.m_bounds, Rgba::YELLOW );
			}
		}
	}
}


//-------------------------------------------------------------------------------------------------
// CheckForCollisionBulletVsBase
// 
// FIXME: we should parameterize collision checks over normalized frame time t:[0,1]
//	in order to handle cases where the bullet would theoretically impact multiple
//	different objects during the frame (rather than only impacting the first one hit).
//	Collision primative checks currently detect IF a collision occurred, but not WHEN.
// 
// Return true if bullet should be destroyed, false if not.
//-------------------------------------------------------------------------------------------------
bool TheGame::CheckForCollisionBulletVsBase( Base& base, Projectile& bullet )
{
	int x, y;
	const float bulletRadius = g_particleTypeRadiusTable[ PROJECTILETYPE_PLAYER_BOLT ];

	if( g_debugDrawCollisions )
	{
		DebugDrawCollisionBulletVsBase( base, bullet );
	}

	/// Check for collision with each node in the base
	for( x = 0; x < base.m_width; x ++ )
	{
		for( y = 0; y < base.m_height; y ++ )
		{
			BaseNode& node = base.m_grid[ y ][ x ];

			/// Ignore if not a killable node type
			if( !(node.m_type == NODETYPE_TERMINATOR || node.m_type == NODETYPE_HEART) )
				continue;

			/// Ignore if node bounds and bullet bounds don't overlap
			if( !bullet.m_moveBounds.IsOverlapping( node.m_bounds ) )
				continue;

			/// Calculate the world x,y of the node's center
			const Vector2 nodePos = base.m_pos + CalcOffsetForGridXY( x, y );

			/// Do a precise check to see if bullet actually collides with node
			if( DoMovingDiscsCollide2d( bullet.m_pos, bullet.m_moveVector, bulletRadius, nodePos, NODE_RADIUS ) )
			{
				ResolveCollisionBulletVsNode( base, bullet, node );
				return true;
			}
		}
	}

	/// Check for collision with each pipe (node connector) in the base
	for( x = 0; x < base.m_width; x ++ )
	{
		for( y = 0; y < base.m_height; y ++ )
		{
			BaseNode& node = base.m_grid[ y ][ x ];

			/// Skip this node if it has no connections
			if( node.IsEmpty() )
				continue;

			if( CheckAndResolveCollisionBulletVsPipe( base, bullet, node, HEXDIRECTION_EAST ) )
				return true;

			if( CheckAndResolveCollisionBulletVsPipe( base, bullet, node, HEXDIRECTION_NORTHEAST ) )
				return true;

			if( CheckAndResolveCollisionBulletVsPipe( base, bullet, node, HEXDIRECTION_NORTHWEST ) )
				return true;
		}
	}

	return false;
}


//-------------------------------------------------------------------------------------------------
// CheckAndResolveCollisionBulletVsPipe
//-------------------------------------------------------------------------------------------------
bool TheGame::CheckAndResolveCollisionBulletVsPipe( Base& base, Projectile& bullet, BaseNode& node, HexDirection direction )
{
	BasePipe* pipe = node.m_connections[ direction ];
	if( !pipe )
		return false;

	/// Don't check if the pipe isn't intact
	if( pipe->m_status != PIPESTATUS_INTACT )
		return false;

	/// Trivial reject if bullet and pipe bounds don't overlap
	if( !bullet.m_moveBounds.IsOverlapping( pipe->m_bounds ) )
		return false;

	/// Calculate the world coordinates of the pipe's start and end points
	const Vector2 pipeStart = pipe->m_startNode->m_pos;
	const Vector2 pipeDisplacement = pipe->m_endNode->m_pos - pipeStart;

	/// Do a precise check to see if bullet actually collides with pipe
	LineSegment2 bulletMotion( bullet.m_pos, bullet.m_moveVector );
	LineSegment2 pipeSegment( pipeStart, pipeDisplacement );
	const float distSquared = DistanceSquaredBetweenLineSegments2d( bulletMotion, pipeSegment );
	const float radiiSummed = PIPE_RADIUS + g_particleTypeRadiusTable[ PROJECTILETYPE_PLAYER_BOLT ];
	const float radiiSquared = radiiSummed * radiiSummed;
	if( distSquared >= radiiSquared )
		return false; // no collision

	if( g_debugDrawCollisions )
	{
		glLineWidth( 2.0f );
		RENDERER->DrawAABB2( pipe->m_bounds, Rgba::WHITE );
		RENDERER->DrawAABB2( bullet.m_moveBounds, Rgba::WHITE );
		glLineWidth( 1.0f );
	}

	ResolveCollisionBulletVsPipe( base, bullet, *pipe, pipeSegment );
	return true;
}


//-------------------------------------------------------------------------------------------------
// CheckForBulletCollisionsVersesBase
//-------------------------------------------------------------------------------------------------
void TheGame::CheckForBulletCollisionsVersesBase( Base& base, const AABB2& bulletBounds, int particleType )
{
	/// Check each projectile individually
	list<Projectile*>::iterator iter;
	for( iter = projectileList[ particleType ].begin(); iter != projectileList[ particleType ].end(); iter ++ )
	{
		Projectile* p = *iter;

		/// Check if this projectile's motion AABB overlaps with the base's bounds
		p->UpdateMoveBounds(); // note: this only performs work if the bounds are out of date
		if( !p->m_moveBounds.IsOverlapping( base.m_bounds ) )
			continue;

		if( g_debugDrawCollisions )
		{
			RENDERER->DrawAABB2( p->m_moveBounds, Rgba::BLUE );
			RENDERER->DrawAABB2( base.m_bounds, Rgba::RED );
		}

		/// Check for collision against all nodes and pipes in the base
		bool removeBullet = CheckForCollisionBulletVsBase( base, *p );
		if( removeBullet )
		{
			list<Projectile*>::iterator dead = iter;
			++ iter;
			projectileList[ particleType ].erase( dead );
			delete p;
		}
	}
}


//-------------------------------------------------------------------------------------------------
// CheckForCollisions
//-------------------------------------------------------------------------------------------------
void TheGame::CheckForCollisions()
{
	/// Build an AABB that bounds all current player bullets
	AABB2 playerBulletBounds = BuildPlayerBulletBounds();

	if( g_debugDrawCollisions )
	{
		RENDERER->DrawAABB2( playerBulletBounds, Rgba::DARKPURPLE );
	}

	/// Build a list of bases whose bounds overlap the player bullet bounds
	list<Base*> overlappingBases;
	list<Base*>::iterator iter;
	for( iter = baseList.begin(); iter != baseList.end(); iter ++ )
	{
		Base* base = *iter;
		if( !base->m_isAlive )
		{
			if( g_debugDrawCollisions )
			{
				RENDERER->DrawAABB2( base->m_bounds, Rgba::GREY );
			}
			continue;
		}

		/// Check if the AABB of this base overlaps that of the player's bullets
		if( playerBulletBounds.IsOverlapping( base->m_bounds ) )
		{
			overlappingBases.push_back( base );
			if( g_debugDrawCollisions )
			{
				RENDERER->DrawAABB2( base->m_bounds, Rgba::YELLOW );
			}
		}
		else
		{
			if( g_debugDrawCollisions )
			{
				RENDERER->DrawAABB2( base->m_bounds, Rgba::DARKGREEN );
			}
		}
	}

	/// Iterate through each base whose bounds overlapped with the player bullets' bounds
	for( iter = overlappingBases.begin(); iter != overlappingBases.end(); iter ++ )
	{
		Base* base = *iter;
		CheckForBulletCollisionsVersesBase( *base, playerBulletBounds, PROJECTILETYPE_PLAYER_BOLT );
	}
}

//-------------------------------------------------------------------------------------------------
// BuildPlayerBulletBounds
//-------------------------------------------------------------------------------------------------
AABB2 TheGame::BuildPlayerBulletBounds()
{
	/// Start with a point-singularity bounding box around the player
	AABB2 bounds( g_player.m_pos );

	/// Add each projectile (and its motion this frame!) to the bounds
	list<Projectile*>::iterator iter;
	for( iter = projectileList[ PROJECTILETYPE_PLAYER_BOLT ].begin(); iter != projectileList[ PROJECTILETYPE_PLAYER_BOLT ].end(); iter ++ )
	{
		Projectile* p = *iter;
		bounds.StretchBoundsToIncludePoint( p->m_pos );
		bounds.StretchBoundsToIncludePoint( p->m_endPos );

		if( g_debugDrawCollisions )
		{
			AABB2 temp;
			temp.SetFromTwoPointsPadding( p->m_pos, p->m_endPos, Vector2( g_particleTypeRadiusTable[ PROJECTILETYPE_PLAYER_BOLT ], g_particleTypeRadiusTable[ PROJECTILETYPE_PLAYER_BOLT ] ) );
			RENDERER->DrawAABB2( temp, Rgba::DARKBLUE );
		}
	}

	/// Expand the final box a bit to take bullet radius into account
	float bulletRadius = g_particleTypeRadiusTable[ PROJECTILETYPE_PLAYER_BOLT ];
	bounds.AddPadding( bulletRadius, bulletRadius );

	return bounds;
}


